import logging
from collections import defaultdict
from typing import TYPE_CHECKING, Optional, override

from ai.backend.common.observer.types import AbstractObserver
from ai.backend.common.types import ContainerStatus
from ai.backend.logging.utils import BraceStyleAdapter

if TYPE_CHECKING:
    from ..agent import AbstractAgent


log = BraceStyleAdapter(logging.getLogger(__spec__.name))


PORT_USAGE_THRESHOLD = 4


class HostPortObserver(AbstractObserver):
    _agent: "AbstractAgent"
    _port_unused_counts: defaultdict[int, int]

    def __init__(
        self,
        agent: "AbstractAgent",
    ) -> None:
        self._agent = agent
        self._port_unused_counts: defaultdict[int, int] = defaultdict(int)

    @property
    @override
    def name(self) -> str:
        return "agent_host_port"

    @override
    async def observe(self) -> None:
        containers = await self._agent.enumerate_containers(ContainerStatus.active_set())
        occupied_host_ports: set[int] = set()
        for _, container in containers:
            for container_port in container.ports:
                occupied_host_ports.add(container_port.host_port)

        unused_ports = self._agent.current_used_port_set() - occupied_host_ports
        for previous_unused_port in list(self._port_unused_counts.keys()):
            if previous_unused_port not in unused_ports:
                del self._port_unused_counts[previous_unused_port]

        ports_to_release = set()
        for unused_port in unused_ports:
            self._port_unused_counts[unused_port] += 1
            if self._port_unused_counts[unused_port] >= PORT_USAGE_THRESHOLD:
                ports_to_release.add(unused_port)
                del self._port_unused_counts[unused_port]
        self._agent.release_unused_ports(ports_to_release)

    @override
    def observe_interval(self) -> float:
        return 15.0

    @classmethod
    @override
    def timeout(cls) -> Optional[float]:
        return 10.0

    @override
    async def cleanup(self) -> None:
        pass



import asyncio
from collections.abc import Mapping
from dataclasses import dataclass
from decimal import Decimal
from pathlib import Path
from typing import Any, override

from aiodocker.docker import Docker

from ai.backend.agent.affinity_map import AffinityMap, AffinityPolicy
from ai.backend.agent.exception import UnsupportedResource
from ai.backend.agent.resources import (
    ComputerContext,
    KernelResourceSpec,
    Mount,
    allocate,
    known_slot_types,
)
from ai.backend.agent.types import MountInfo
from ai.backend.agent.utils import update_nested_dict
from ai.backend.common.asyncio import closing_async
from ai.backend.common.stage.types import (
    ArgsSpecGenerator,
    Provisioner,
    ProvisionStage,
)
from ai.backend.common.types import (
    DeviceName,
    MountPermission,
    ResourceSlot,
    current_resource_slots,
)


@dataclass
class ResourceSpec:
    """
    Specification for resource allocation stage.

    This includes the requested resource slots and all necessary context
    for allocating compute resources including accelerators.
    """

    resource_slots: ResourceSlot
    resource_opts: Mapping[str, Any]

    computers: Mapping[DeviceName, ComputerContext]
    allocation_order: list[DeviceName]
    affinity_map: AffinityMap
    affinity_policy: AffinityPolicy
    allow_fractional_resource_fragmentation: bool

    config_dir: Path


class ResourceSpecGenerator(ArgsSpecGenerator[ResourceSpec]):
    pass


@dataclass
class ResourceResult:
    resource_spec: KernelResourceSpec
    resource_opts: Mapping[str, Any]
    accelerator_mounts: list[Mount]
    container_arg: dict[str, Any]


class ResourceProvisioner(Provisioner[ResourceSpec, ResourceResult]):
    """
    Provisioner for the kernel creation setup stage.
    This provisioner handles resource allocation including accelerators (GPU, TPU, etc.).
    """

    def __init__(
        self,
        resource_lock: asyncio.Lock,
    ) -> None:
        self._lock = resource_lock

    @property
    @override
    def name(self) -> str:
        return "docker-resource"

    @override
    async def setup(self, spec: ResourceSpec) -> ResourceResult:
        resource_spec, resource_opts = self._prepare_resource_spec(spec)
        await self._allocate(spec, resource_spec)

        # Generate accelerator mounts for all allocated devices
        accelerator_mounts = await self._prepare_accelerator_mounts(spec, resource_spec)

        # Apply accelerator-specific configurations and collect environment variables
        container_arg = await self._prepare_accelerator_configs(spec, resource_spec)

        return ResourceResult(
            resource_spec=resource_spec,
            resource_opts=resource_opts,
            accelerator_mounts=accelerator_mounts,
            container_arg=container_arg,
        )

    def _prepare_resource_spec(
        self, spec: ResourceSpec
    ) -> tuple[KernelResourceSpec, Mapping[str, Any]]:
        slots = spec.resource_slots
        # Ensure that we have intrinsic slots.
        # But why do we need assertions here?
        # assert SlotName("cpu") in slots
        # assert SlotName("mem") in slots

        # accept unknown slot type with zero values
        # but reject if they have non-zero values.
        for st, sv in slots.items():
            if st not in known_slot_types and sv != Decimal(0):
                raise UnsupportedResource(st)
        # sanitize the slots
        current_resource_slots.set(known_slot_types)
        slots = slots.normalize_slots(ignore_unknown=True)
        resource_spec = KernelResourceSpec(
            allocations={},
            slots={**slots},  # copy
            mounts=[],
            scratch_disk_size=0,  # TODO: implement (#70)
        )
        resource_opts = spec.resource_opts
        return resource_spec, resource_opts

    async def _allocate(self, spec: ResourceSpec, kernel_resource_spec: KernelResourceSpec) -> None:
        async with self._lock:
            allocate(
                spec.computers,
                kernel_resource_spec,
                spec.allocation_order,
                spec.affinity_map,
                spec.affinity_policy,
                allow_fractional_resource_fragmentation=spec.allow_fractional_resource_fragmentation,
            )

    async def _prepare_accelerator_mounts(
        self, spec: ResourceSpec, resource_spec: KernelResourceSpec
    ) -> list[Mount]:
        mounts: list[MountInfo] = []

        for dev_type, device_alloc in resource_spec.allocations.items():
            if dev_type not in spec.computers:
                continue

            computer_ctx = spec.computers[dev_type]
            computer = computer_ctx.instance

            # Create directory for accelerator-specific files
            src_path = spec.config_dir / str(computer.key)
            src_path.mkdir(exist_ok=True)

            # Generate mounts for this accelerator type
            accelerator_mounts = await computer.generate_mounts(src_path, device_alloc)
            mounts.extend(accelerator_mounts)

        return [
            Mount(
                m.mode,
                m.src_path,
                Path(m.dst_path.as_posix()),
                MountPermission.READ_WRITE,  # Accelerators typically need read-write access
            )
            for m in mounts
        ]

    async def _prepare_accelerator_configs(
        self, spec: ResourceSpec, resource_spec: KernelResourceSpec
    ) -> dict[str, Any]:
        container_arg: dict[str, Any] = {}

        async with closing_async(Docker()) as docker:
            for dev_type, device_alloc in resource_spec.allocations.items():
                computer_ctx = spec.computers[dev_type]
                computer = computer_ctx.instance

                update_nested_dict(
                    container_arg,
                    await computer.generate_docker_args(docker, device_alloc),
                )

        return container_arg

    @override
    async def teardown(self, resource: ResourceResult) -> None:
        """
        Release allocated resources.
        """
        # TODO: Implement proper resource deallocation
        # This should deallocate resources from the computers' allocation maps
        pass


class ResourceStage(ProvisionStage[ResourceSpec, ResourceResult]):
    """
    Stage for resource allocation in the kernel lifecycle.

    This stage handles:
    1. Allocation of compute resources (CPU, memory, etc.)
    2. Allocation of accelerator resources (GPU, TPU, etc.)
    3. Generation of accelerator-specific mounts
    4. Application of accelerator-specific configurations

    The accelerator_mounts generated by this stage should be used by
    KernelRunnerMountStage to properly mount accelerator files/directories.

    Example usage:
        # In the kernel creation pipeline
        resource_provisioner = ResourceProvisioner(resource_lock, computers, agent)
        resource_stage = ResourceStage(resource_provisioner)

        # After stage execution
        result = await resource_stage.wait_for_resource()
        # result.accelerator_mounts contains all accelerator-specific mounts
    """

    pass



from http import HTTPStatus
from typing import Any, Mapping, Sequence

from aiohttp import web

from ai.backend.agent.docker.kernel import DockerKernel

from .plugin import MetadataPlugin, MetadataPluginRoute, NewMetadataPluginResponse


class ContainerMetadataPlugin(MetadataPlugin):
    async def init(self, context):
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, plugin_config: Mapping[str, Any]) -> None:
        pass

    async def prepare_app(self) -> NewMetadataPluginResponse:
        app = web.Application()
        return NewMetadataPluginResponse(app, [])

    async def routes(self) -> Sequence[MetadataPluginRoute]:
        return [
            MetadataPluginRoute("GET", "/envs", self.get_envs, None),
            MetadataPluginRoute("GET", "/sandbox", self.get_sandbox_type, None),
            MetadataPluginRoute("GET", "/local-ipv4", self.get_local_ipv4, None),
        ]

    async def get_envs(self, request: web.Request) -> web.Response:
        kernel: DockerKernel = request["kernel"]
        if kernel is None:
            return web.Response(status=HTTPStatus.NOT_FOUND)
        response = dict(kernel.environ)
        return web.json_response(response)

    async def get_local_ipv4(self, request: web.Request) -> web.Response:
        return web.Response(body=request["container-ip"])

    async def get_sandbox_type(self, request: web.Request) -> web.Response:
        return web.Response(body=self.local_config["container"]["sandbox-type"])




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/agent:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.13',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    'description': 'Backend.AI Agent',
    'entry_points': {
        'backendai_cli_v10': [
            'ag = ai.backend.agent.cli.__main__:main',
            'ag.start-server = ai.backend.agent.server:main',
        ],
        'backendai_network_agent_v1': [
            'overlay = ai.backend.agent.docker.intrinsic:OverlayNetworkPlugin',
            'host = ai.backend.agent.docker.intrinsic:HostNetworkPlugin',
        ],
    },
    'install_requires': (
        'aiodocker==0.24.0',
        'aiofiles~=24.1.0',
        'aiohttp_cors~=0.8.1',
        'aiohttp~=3.12.15',
        'aiomonitor~=0.7.0',
        'aiotools~=1.9.0',
        'async_timeout~=4.0',
        'attrs>=25.3',
        """backend.ai-cli==25.15.3
""",
        """backend.ai-common==25.15.3
""",
        """backend.ai-kernel-binary==25.15.3
""",
        """backend.ai-kernel-helper==25.15.3
""",
        """backend.ai-kernel==25.15.3
""",
        'backend.ai-krunner-static-gnu==4.4.0',
        """backend.ai-logging==25.15.3
""",
        """backend.ai-plugin==25.15.3
""",
        'cachetools~=5.5.0',
        'callosum~=1.0.3',
        'cattrs~=24.1.1',
        'click~=8.1.7',
        'etcd-client-py~=0.4.1',
        'janus~=2.0',
        'kubernetes-asyncio~=33.3.0',
        'kubernetes~=33.1.0',
        'more-itertools~=10.5.0',
        'networkx~=3.3.0',
        'prometheus-client~=0.21.1',
        'psutil~=7.0',
        'pydantic[email]~=2.11.3',
        'pyzmq~=26.4',
        'ruamel.yaml~=0.18.10',
        'setproctitle~=1.3.5',
        'setuptools~=80.0.0',
        'tenacity>=9.0',
        'tomlkit~=0.13.2',
        'trafaret~=2.1',
        'types-aiofiles',
        'types-cachetools',
        'types-psutil',
        'types-setuptools',
        'typing_extensions~=4.11',
        'uvloop~=0.21; sys_platform != "Windows"',
    ),
    'license': 'LGPLv3',
    'long_description': """# Backend.AI Agent

The Backend.AI Agent is a small daemon that does:

* Reports the status and available resource slots of a worker to the manager
* Routes code execution requests to the designated kernel container
* Manages the lifecycle of kernel containers (create/monitor/destroy them)

## Package Structure

* `ai.backend`
  - `agent`: The agent package
    - `docker`: A docker-based backend implementation for the kernel lifecycle interface.
    - `server`: The agent daemon which communicates with the manager and the Docker daemon
    - `watcher`: A side-by-side daemon which provides a separate HTTP endpoint for accessing the status
      information of the agent daemon and manipulation of the agent's systemd service
  - `helpers`: A utility package that is available as `ai.backend.helpers` *inside* Python-based containers
  - `kernel`: Language-specific runtimes (mostly ipykernel client adaptor) which run *inside* containers
  - `runner`: Auxiliary components (usually self-contained binaries) mounted *inside* containers


## Installation

Please visit [the installation guides](https://github.com/lablup/backend.ai/wiki).


### Kernel/system configuration

#### Recommended kernel parameters in the bootloader (e.g., Grub):

```
cgroup_enable=memory swapaccount=1
```

#### Recommended resource limits:

**`/etc/security/limits.conf`**
```
root hard nofile 512000
root soft nofile 512000
root hard nproc 65536
root soft nproc 65536
user hard nofile 512000
user soft nofile 512000
user hard nproc 65536
user soft nproc 65536
```

**sysctl**
```
fs.file-max=2048000
fs.inotify.max_user_watches=524288
net.core.somaxconn=1024
net.ipv4.tcp_max_syn_backlog=1024
net.ipv4.tcp_slow_start_after_idle=0
net.ipv4.tcp_fin_timeout=10
net.ipv4.tcp_window_scaling=1
net.ipv4.tcp_tw_reuse=1
net.ipv4.tcp_early_retrans=1
net.ipv4.ip_local_port_range=40000 65000
net.core.rmem_max=16777216
net.core.wmem_max=16777216
net.ipv4.tcp_rmem=4096 12582912 16777216
net.ipv4.tcp_wmem=4096 12582912 16777216
net.netfilter.nf_conntrack_max=10485760
net.netfilter.nf_conntrack_tcp_timeout_established=432000
net.netfilter.nf_conntrack_tcp_timeout_close_wait=10
net.netfilter.nf_conntrack_tcp_timeout_fin_wait=10
net.netfilter.nf_conntrack_tcp_timeout_time_wait=10
```

The `ip_local_port_range` should not overlap with the container port range pool
(default: 30000 to 31000).

To apply netfilter settings during the boot time, you may need to add `nf_conntrack` to `/etc/modules`
so that `sysctl` could set the `net.netfilter.nf_conntrack_*` values.


### For development

#### Prerequisites

* Python 3.6 or higher with [pyenv](https://github.com/pyenv/pyenv)
and [pyenv-virtualenv](https://github.com/pyenv/pyenv-virtualenv) (optional but recommneded)
* Docker 18.03 or later with docker-compose (18.09 or later is recommended)

First, you need **a working manager installation**.
For the detailed instructions on installing the manager, please refer
[the manager's README](https://github.com/lablup/backend.ai-manager/blob/master/README.md)
and come back here again.

#### Preparing working copy

Install and activate [`git-lfs`](https://git-lfs.github.com/) to work with pre-built binaries in
`src/ai/backend/runner`.

```console
$ git lfs install
```

Next, prepare the source clone of the agent and install from it as follows.
`pyenv` is just a recommendation; you may use other virtualenv management tools.

```console
$ git clone https://github.com/lablup/backend.ai-agent agent
$ cd agent
$ pyenv virtualenv venv-agent
$ pyenv local venv-agent
$ pip install -U pip setuptools
$ pip install -U -r requirements/dev.txt
```

### Linting

We use `flake8` and `mypy` to statically check our code styles and type consistency.
Enable those linters in your favorite IDE or editor.

### Halfstack (single-node development & testing)

With the halfstack, you can run the agent simply.
Note that you need a working manager running with the halfstack already!

#### Recommended directory structure

* `backend.ai-dev`
  - `manager` (git clone from [the manager repo](https://github.com/lablup/backend.ai-manager))
  - `agent` (git clone from here)
  - `common` (git clone from [the common repo](https://github.com/lablup/backend.ai-common))

Install `backend.ai-common` as an editable package in the agent (and the manager) virtualenvs
to keep the codebase up-to-date.

```console
$ cd agent
$ pip install -U -e ../common
```

#### Steps

```console
$ mkdir -p "./scratches"
$ cp config/halfstack.toml ./agent.toml
```

If you're running agent under linux, make sure you've set appropriate iptables rule 
before starting agent. This can be done by executing script `scripts/update-metadata-iptables.sh` 
before each agent start.

Then, run it (for debugging, append a `--debug` flag):

```console
$ python -m ai.backend.agent.server
```

To run the agent-watcher:

```console
$ python -m ai.backend.agent.watcher
```

The watcher shares the same configuration TOML file with the agent.
Note that the watcher is only meaningful if the agent is installed as a systemd service
named `backendai-agent.service`.

To run tests:

```console
$ python -m flake8 src tests
$ python -m pytest -m 'not integration' tests
```


## Deployment

### Configuration

Put a TOML-formatted agent configuration (see the sample in `config/sample.toml`)
in one of the following locations:

 * `agent.toml` (current working directory)
 * `~/.config/backend.ai/agent.toml` (user-config directory)
 * `/etc/backend.ai/agent.toml` (system-config directory)

Only the first found one is used by the daemon.

The agent reads most other configurations from the etcd v3 server where the cluster
administrator or the Backend.AI manager stores all the necessary settings.

The etcd address and namespace must match with the manager to make the agent
paired and activated.
By specifying distinguished namespaces, you may share a single etcd cluster with multiple
separate Backend.AI clusters.

By default the agent uses `/var/cache/scratches` directory for making temporary
home directories used by kernel containers (the `/home/work` volume mounted in
containers).  Note that the directory must exist in prior and the agent-running
user must have ownership of it.  You can change the location by
`scratch-root` option in `agent.toml`.

### Running from a command line

The minimal command to execute:

```sh
python -m ai.backend.agent.server
python -m ai.backend.agent.watcher
```

For more arguments and options, run the command with `--help` option.

### Example config for systemd

`/etc/systemd/system/backendai-agent.service`:

```dosini
[Unit]
Description=Backend.AI Agent
Requires=docker.service
After=network.target remote-fs.target docker.service

[Service]
Type=simple
User=root
Group=root
Environment=HOME=/home/user
ExecStart=/home/user/backend.ai/agent/run-agent.sh
WorkingDirectory=/home/user/backend.ai/agent
KillMode=process
KillSignal=SIGTERM
PrivateTmp=false
Restart=on-failure
RestartSec=5

[Install]
WantedBy=multi-user.target
```

`/home/user/backend.ai/agent/run-agent.sh`:

```sh
#! /bin/sh
if [ -z "$PYENV_ROOT" ]; then
  export PYENV_ROOT="$HOME/.pyenv"
  export PATH="$PYENV_ROOT/bin:$PATH"
fi
eval "$(pyenv init -)"
eval "$(pyenv virtualenv-init -)"

cd /home/user/backend.ai/agent
if [ "$#" -eq 0 ]; then
  sh /home/user/backend.ai/agent/scripts/update-metadata-iptables.sh
  exec python -m ai.backend.agent.server
else
  exec "$@"
fi
```

### Networking

The manager and agent should run in the same local network or different
networks reachable via VPNs, whereas the manager's API service must be exposed to
the public network or another private network that users have access to.

The manager must be able to access TCP ports 6001, 6009, and 30000 to 31000 of the agents in default
configurations.  You can of course change those port numbers and ranges in the configuration.

| Manager-to-Agent TCP Ports | Usage |
|:--------------------------:|-------|
| 6001                       | ZeroMQ-based RPC calls from managers to agents |
| 6009                       | HTTP watcher API |
| 30000-31000                | Port pool for in-container services |

The operation of agent itself does not require both incoming/outgoing access to
the public Internet, but if the user's computation programs need the Internet, the docker containers
should be able to access the public Internet (maybe via some corporate firewalls).

| Agent-to-X TCP Ports     | Usage |
|:------------------------:|-------|
| manager:5002             | ZeroMQ-based event push from agents to the manager |
| etcd:2379                | etcd API access |
| redis:6379               | Redis API access |
| docker-registry:{80,443} | HTTP watcher API |
| (Other hosts)            | Depending on user program requirements |


LICENSES
--------

[GNU Lesser General Public License](https://github.com/lablup/backend.ai-agent/blob/master/LICENSE)
[Dependencies](https://github.com/lablup/backend.ai-manager/blob/agent/DEPENDENCIES.md)
""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-agent',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.agent': (
            'VERSION',
            'py.typed',
        ),
        'ai.backend.agent.docker': (
            'backendai-socket-relay.img.aarch64.tar.gz',
            'backendai-socket-relay.img.x86_64.tar.gz',
            'backendai-socket-relay.version.txt',
            'linuxkit-metadata-proxy-worker.aarch64.bin',
            'linuxkit-metadata-proxy-worker.x86_64.bin',
            'linuxkit-nsenter.img.aarch64.tar.gz',
            'linuxkit-nsenter.img.x86_64.tar.gz',
            'linuxkit-nsenter.version.txt',
        ),
    },
    'packages': (
        'ai.backend.agent',
        'ai.backend.agent.cli',
        'ai.backend.agent.config',
        'ai.backend.agent.docker',
        'ai.backend.agent.docker.metadata',
        'ai.backend.agent.dummy',
        'ai.backend.agent.kubernetes',
        'ai.backend.agent.metrics',
        'ai.backend.agent.observer',
        'ai.backend.agent.plugin',
        'ai.backend.agent.stage',
        'ai.backend.agent.stage.kernel_lifecycle',
        'ai.backend.agent.stage.kernel_lifecycle.docker',
        'ai.backend.agent.stage.kernel_lifecycle.docker.mount',
        'ai.backend.agent.vendor',
        'ai.backend.agent.watcher',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.13,<3.14',
    'url': 'https://github.com/lablup/backend.ai',
    'version': """25.15.3
""",
    'zip_safe': False,
})



import base64
from dataclasses import dataclass
from http import HTTPStatus
from typing import Optional, override

from aiodocker.docker import Docker
from aiodocker.exceptions import DockerError

from ai.backend.common.asyncio import closing_async
from ai.backend.common.docker import ImageRef
from ai.backend.common.exception import ImageNotAvailable
from ai.backend.common.stage.types import (
    Provisioner,
    ProvisionStage,
    SpecGenerator,
)
from ai.backend.common.types import AutoPullBehavior, ImageRegistry


@dataclass
class ImagePullCheckSpec:
    image_ref: ImageRef
    image_digest: str
    registry_conf: ImageRegistry
    pull_timeout: Optional[float]
    auto_pull_behavior: AutoPullBehavior


@dataclass
class ImagePullSpec:
    do_pull: bool
    image_ref: ImageRef
    registry_conf: ImageRegistry
    pull_timeout: Optional[float]


class ImagePullSpecProvisioner(Provisioner[ImagePullCheckSpec, ImagePullSpec]):
    @property
    def name(self) -> str:
        return "docker-image-pull-check"

    async def setup(self, spec: ImagePullCheckSpec) -> ImagePullSpec:
        do_pull = (not spec.image_ref.is_local) and await self._check_image_exist(spec)
        return ImagePullSpec(
            do_pull=do_pull,
            image_ref=spec.image_ref,
            registry_conf=spec.registry_conf,
            pull_timeout=spec.pull_timeout,
        )

    async def _check_image_exist(self, spec: ImagePullCheckSpec) -> bool:
        try:
            async with closing_async(Docker()) as docker:
                image_info = await docker.images.inspect(spec.image_ref.canonical)
                if spec.auto_pull_behavior == AutoPullBehavior.DIGEST:
                    if image_info["Id"] != spec.image_digest:
                        return True
        except DockerError as e:
            if e.status == HTTPStatus.NOT_FOUND:
                match spec.auto_pull_behavior:
                    case AutoPullBehavior.DIGEST | AutoPullBehavior.TAG:
                        return True
                    case AutoPullBehavior.NONE:
                        raise ImageNotAvailable(spec.image_ref)
            else:
                raise
        return False

    async def teardown(self, resource: ImagePullSpec) -> None:
        pass


class ImagePullSpecGenerator(SpecGenerator[ImagePullSpec]):
    def __init__(self, args: ImagePullCheckSpec) -> None:
        self._args = args
        self._provisioner = ImagePullSpecProvisioner()

    @override
    async def wait_for_spec(self) -> ImagePullSpec:
        return await self._provisioner.setup(self._args)


@dataclass
class ImagePullResult:
    image_ref: ImageRef
    did_pull: bool


class ImagePullProvisioner(Provisioner[ImagePullSpec, ImagePullResult]):
    @property
    @override
    def name(self) -> str:
        return "docker-image-pull"

    @override
    async def setup(self, spec: ImagePullSpec) -> ImagePullResult:
        if spec.do_pull:
            image_ref = await self._pull_image(spec)
            did_pull = True
        else:
            image_ref = spec.image_ref
            did_pull = False
        return ImagePullResult(image_ref=image_ref, did_pull=did_pull)

    async def _pull_image(self, spec: ImagePullSpec) -> ImageRef:
        auth_config: Optional[dict[str, str]] = None
        reg_user = spec.registry_conf.get("username")
        reg_passwd = spec.registry_conf.get("password")
        if reg_user is not None and reg_passwd is not None:
            encoded_creds = base64.b64encode(f"{reg_user}:{reg_passwd}".encode("utf-8")).decode(
                "ascii"
            )
            auth_config = {
                "auth": encoded_creds,
            }
        async with closing_async(Docker()) as docker:
            result = await docker.images.pull(
                spec.image_ref.canonical, auth=auth_config, timeout=spec.pull_timeout
            )

            if not result:
                raise RuntimeError("Failed to pull image: unexpected return value from aiodocker")
            elif error := result[-1].get("error"):
                raise RuntimeError(f"Failed to pull image: {error}")

        return spec.image_ref

    @override
    async def teardown(self, resource: ImagePullResult) -> None:
        pass


class ImagePullStage(ProvisionStage[ImagePullSpec, ImagePullResult]):
    pass
